#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>

#include <rtai.h>
#include <rtai_sched.h>
#include <rtai_shm.h>
#include <rtai_nam2num.h>

#include <rtai_math.h>

#define STACK_SIZE    4000

#include "adc.h"
#include "hinf.h"
#include "common.h"

static RT_TASK thread;

static float * time_resp;
static double * cont_real;
static struct my_msg_struct * msg;
static int *pos_ref;
static int *flag;

static void fun(long t_dummy)
{
	RTIME s_time;
	float t, t0;
	int count_record;

	int i, j, k, failed = 0;
	double y, w, theta;
	volatile double *p;

	double x; // reference for linear motor's position 
	double xk[ORDER_MAX]; // controller's states
	double xk_tmp[ORDER_MAX];

	while(1){
		int ret;
		int err;

		if (msg->command == START_TASK) {
			msg->command = STOP_TASK;
			*flag = 1;
			*pos_ref = PULSE_OFFSET;
			
			count_record = 0;
			x = 0.;
			for(i = 0; i < msg->order; i++){
				xk[i] = 0.;
			}

			s_time = rt_get_time();
			rt_task_make_periodic_relative_ns(&thread, 10000000, msg->sampling_period*1e9);
		}

		if(*flag == 1){
			  
			t = count2nano(rt_get_time() - s_time)*1.0e-9;

			//			if(t < 0.1){
			//				*pos_ref = PULSE_OFFSET + DIST_AMP*X2PULSE;
			//			}else{
			//				*pos_ref = PULSE_OFFSET;
			//			}

			ad_conv(&y);
			theta = (y - V_OFFSET) * V2THETA;
			if((y > V_MAX)||(y < V_MIN)){
				failed = 1;
			}

			if((count_record > BUF_LEN * WAITING_SAMPLES)&&(failed == 0)){
				x = 0.;
				p = &cont_real[msg->order * (msg->order + 1)]; // move to C matrix
				for(i = 0; i < msg->order; i++){
					x += *p * xk[i];
					p++;
				}
				x += *p * theta;
				
				if(x > XMAX) x = XMAX;
				if(x < XMIN) x = XMIN;
#ifndef NO_CONTROL
				*pos_ref = PULSE_OFFSET + (int)(x * X2PULSE);
#endif

				p = cont_real;
				for(i = 0; i < msg->order; i++){
					xk_tmp[i] = 0.;
					for(j = 0; j < msg->order; j++){
						xk_tmp[i] += *p * xk[j];
						p++;
					}
					xk_tmp[i] += *p * theta;
					p++;
				}
				memcpy(xk, xk_tmp, sizeof(xk));
			}
		}
		
		if(count_record < BUF_LEN * RECORDING_SAMPLES){
			time_resp[count_record++] = t;
			time_resp[count_record++] = y;
			time_resp[count_record++] = theta;
			time_resp[count_record++] = x;
			time_resp[count_record++] = *pos_ref;
			//			time_resp[count_record++] = xk[0];
			//			time_resp[count_record++] = xk[1];
		}else{
			*flag = 0;
			rt_task_make_periodic_relative_ns(&thread, 10000000, 10000000);
		}
		
		rt_task_wait_period();

	}
}

int
init_module(void)
{
	init_adc();

        if((flag = (int *)rtai_kmalloc(nam2num("flag"), sizeof(int))) ==NULL){
                return -ENOMEM;
        }

	if((pos_ref = (int *)rtai_kmalloc(nam2num("pos_ref"), sizeof(int))) ==NULL){
		return -ENOMEM;
	}

	if((time_resp = (float *)rtai_kmalloc(nam2num("time_resp"), BUF_LEN * RECORDING_SAMPLES * sizeof(float))) == NULL){
		return -ENOMEM;
	}

	if((cont_real = (double *)rtai_kmalloc(nam2num("cont_real"), (ORDER_MAX + 1) * (ORDER_MAX + 1) * sizeof(double))) == NULL){
		return -ENOMEM;
	}

	if((msg = (struct my_msg_struct *)rtai_kmalloc(nam2num("msg"), sizeof(struct my_msg_struct))) == NULL){
		return -ENOMEM;
	}

	msg->command = STOP_TASK;
	*flag = 0;

	rt_task_init(&thread, fun, 0, STACK_SIZE, 1, 1, 0);
	rt_linux_use_fpu(1);
	rt_set_oneshot_mode();
	start_rt_timer(0);
	rt_task_make_periodic_relative_ns(&thread, 10000000, 10000000);

	return 0;
}

void
cleanup_module(void)
{
	rt_task_delete(&thread);

	rtai_kfree(nam2num("time_resp"));
	rtai_kfree(nam2num("cont_real"));
	rtai_kfree(nam2num("msg"));
	rtai_kfree(nam2num("pos_ref"));
	rtai_kfree(nam2num("flag"));
	
	cleanup_adc();
}	
